-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

log = function(...) Lua:log(...) end
print = function(...)
    local args = { n = select("#", ...), ... }
    local s_args = {}
    for i = 1, args.n do
        table.insert(s_args, tostring(args[i]))
    end
    Lua:log('A', "print", table.concat(s_args, ', ')) 
end

package.path = "lua/t3d/?.lua;lua/gui/?.lua;lua/common/?.lua;lua/common/socket/?.lua;lua/?.lua;?.lua"
require("utils")

t3d = require("t3d_utils")
json = require("json")
map = require("map")
bindings = require("input_bindings")
guihooks = require("guihooks")
screenshot = require("screenshot")
virtualinput = require("virtualinput")
environment = require("environment")
extensions = require("extensions")
settings = require("settings")

perf = require("perf")

-- automatic loading disabled for runtime performance reasons
extensions.addModulePath("lua/t3d/extensions/")
--extensions.loadModules("lua/t3d/extensions/")

--Lua:enableStackTraceFile("lua.t3d.stack.txt", true)

-- rewrite some functions for T3D usage
-- TODO: clean these up
logAlways=print

-- improve stacktraces
local STP = require "StackTracePlus"
debug.traceback = STP.stacktrace


function requestGameState()
    local level = getTSVar('$Server::MissionFile')
    local state = 'menu'
    if level ~= '' then
        state = 'play'
    end
    be:executeJS('HookManager.trigger("GameStateUpdate","'..state..'")')
end

function clientStartMission(mission)
    --print("starting mission: " .. tostring(mission))
    map.onClientStartMission()
    environment.reset()
    requestGameState()
    extensions.hook('onClientStartMission', mission)
end

function clientEndMission(mission)
    requestGameState()
    extensions.hook('onClientEndMission', mission)
end



function luaPreRender(dt)
    map.updateGFX(dt)
    map.drawDebug(dt, Lua.lastDebugFocusPos)
    extensions.hook('onPreRender', dt)
    extensions.hook('onDrawDebug', Lua.lastDebugFocusPos)
end

gAutoHideDashboard = true
local firstUpdate = true
function update()
    if firstUpdate then
        firstUpdate = false

        reloadAllBindings()

        log('D', "update", " == INIT == ")
        virtualinput.init()

        extensions.hook('onFirstUpdate')
        settings.onFirstUpdate()
    end
    
    --local kB = Lua:getMemoryUsageKB()
    --log('D', "update", "Lua memory usage: " .. kB .. "kB")
   
    virtualinput.update()
    screenshot.updateGFX()

    extensions.hook('onUpdate')
    perf.update()

    if gAutoHideDashboard and Platform.getRealMilliseconds() - getCEFFocusMouseLastTimeMs() > 15000 then
        guihooks.trigger('MenuHide')
    end 
end

function init()
    settings.init()
    guihooks.trigger("EngineLuaReloaded")
    log('D', "init", 'T3D Lua working')

    environment.reset()


    -- hardcoded extensions that are always loaded
    extensions.loadModule('scenarios')
	extensions.loadModule('vehicles')
    extensions.loadModule('campaigns')
    extensions.loadModule('modmanager')
    extensions.loadModule('hardwareinfo')
    
    -- uncomment this to enable the remote controllers
    --extensions.loadModule('remoteController')
    
    extensions.hook('onInit')    
end

function onBeamNGWaypoint(args)
    map.onWaypoint(args)
    extensions.hook('onBeamNGWaypoint', args)
end

function onFileChanged(filename, type)
    --print("onFileChanged: " .. tostring(filename) .. ' : ' .. tostring(type))
    settings.onFileChanged(filename, type)
    map.onFileChanged(filename, type)
    extensions.hook('onFileChanged', filename, type)

    if string.startswith(filename, 'html/') then
        guihooks.trigger('FileChanged', {filename = filename, type = type})
    end
end


function getLevelList()
    if not FS:directoryExists('levels/') then
        return {}
    end
    local files = FS:findFilesByPattern('game:levels/', '*.mis', true, true, false)
    
    -- filter paths to only return filename without extension
    for k,v in pairs(files) do
        files[k] = string.gsub(files[k], "(game:levels/)(.*)%.mis", "%2")
    end
    --dump( files )
    return files
end

function physicsEngineEvent(...)
    local args = unpack({...})
    extensions.hook('onPhysicsEngineEvent', args)
end



-- input things
-- TODO: cleanup

function removeFile(path)
    local success, reason = os.remove(path)
    if not success then
        log('E', '', "Error removing file "..path..": "..reason); end
end
function reloadAllBindings()
    bindings.reloadAllBindings()
end
function saveBindingsToDisk(path, data)
    local success, jsondata = pcall(json.decode, data)
    if not success then
        log('E', '', "Error parsing bindings for file "..path..": "..jsondata); return false end
    bindings.saveBindingsToDisk(path, jsondata)
end
function sendBindingsToGUI()
    be:executeJS("HookManager.trigger('ActionCategoriesChanged', "..encodeJson(bindings.getActionCategories())..");")
    be:executeJS("HookManager.trigger('ActionsChanged', "..encodeJson(bindings.getActions())..");")
    local result = {}
    for devname,inputmapPath in pairs(bindings.getInputmapPaths(bindings.devices)) do
        contents = bindings.readBindingsFromDisk(inputmapPath)
        if contents == false then
            log('E', '', "Error reading bindings from file "..inputmapPath)
        else
            table.insert(result, {inputmapPath = inputmapPath, devname = devname, contents = contents})
        end
    end
    local bindingsJson = encodeJson(result)
    be:executeJS("HookManager.trigger('BindingsChanged', "..bindingsJson..");")
    be:executeJS("HookManager.trigger('BindingTemplateChanged', "..encodeJson(bindings.fillNormalizeBindingDefaults({}))..");")
end
